# Game Raccoon Revision 0 Source Code
# Copyright (C) 2019 Mathew Carr

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License Version 3
# with Additional Terms as found in the file
# LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
##

# The purpose of this file is to work around deficiencies in the mot syntax of 
# vasm and help produce a final Game Raccoon ROM that can be booted from $200
# but instantly copied to $00FF0000 and run from there.

import os
import argparse

# Give a list of binaries to include here (resources, etc.),
# and the Python code will 
# generate an included_binaries.bin which can be included
# by the entry_point.68k.asm outer layer, and an
# included_binaries.68k.asm which can be included by 
# raccoonramresident.68k.asm to give the ROM addresses of the
# included binaries, given that they're going to be present
# in the ROM Boot partition starting at $200.
INCLUDED_BINARIES = {
   "parent_text_nullt_string.bin": "string_parent",
   "TrioDX_font_1bit.bin":         "binary_triodx_font_1bit_bin"
}

if __name__ == "__main__":
   parser = argparse.ArgumentParser()
   parser.add_argument("stage")
   args = parser.parse_args()
   stage = args.stage
   
   # fruitsalad stages are like Makefile targets except I say so.
   
   # Make included_binaries.bin containing included binaries contiguously and longword aligned.
   # Make included_binaries.68k.asm containing the absolute addresses of
   # these binaries when they're incbinned into the final Game Raccoon ROM
   # at address $200 after the Mega Drive header.
   if stage == "generate_included_binaries_blob":
      # Address accumulates as we add files to the blob.
      accumulated_address = 0x200
      with open('included_binaries.bin', 'wb') as included_binaries_bin_file:
         with open('included_binaries.68k.asm', 'w') as included_binaries_68k_asm_file:
            for ib_filename, symbol_name in INCLUDED_BINARIES.items():
               # Pad files to 4-byte boundary with zeroes.
               readbytes = open(ib_filename, 'rb').read()
               readbytes += b'\x00'*3
               readbytes = readbytes[:len(readbytes) & ~3]
               binary_length = len(readbytes)
               
               address_s = accumulated_address               
               accumulated_address += binary_length
               address_e = accumulated_address
               
               included_binaries_bin_file.write(readbytes)
               included_binaries_68k_asm_file.write("{symbol: <50} = ${address:08X}\n".format(symbol=symbol_name                , address=address_s          ))
               included_binaries_68k_asm_file.write("{symbol: <50} = ${address:08X}\n".format(symbol=symbol_name+"_end"         , address=address_e          ))
               included_binaries_68k_asm_file.write("{symbol: <50} = ${length:08X}\n".format( symbol=symbol_name+"_length_bytes", length =binary_length      ))
               included_binaries_68k_asm_file.write("{symbol: <50} = ${length:08X}\n".format( symbol=symbol_name+"_length_words", length =binary_length//2   ))
               included_binaries_68k_asm_file.write("{symbol: <50} = ${length:08X}\n".format( symbol=symbol_name+"_length_longs", length =binary_length//4   ))
               included_binaries_68k_asm_file.write("\n")
               
      exit(0)
            
   # Assemble the Game Raccoon main RAM-resident code based at $00FF0000
   # into a separate file. vasm doesn't like it when you have non-contiguous
   # assembler destination addresses (say, if you had a Mega Drive header at $00000000
   # and a RAM-resident part at $00FF0000): it makes tons of annoying padding.
   # This way we can assemble just the $00FF0000-> block by itself.
   if stage == "compile_ramresident":
      os.system('vasmm68k_mot_win32.exe -Fbin -o raccoonramresident.bin raccoonramresident.68k.asm')
      exit(0)
         
   # Create the game raccoon boot image file containing the entry point stub,
   # the incbinned binaries, and the incbinned .bin RAM-resident part all together.
   # This is the complete Game Raccoon boot image, except without any directory cache present.
   if stage == "combine_entry_point_ramresident":
      os.system('vasmm68k_mot_win32.exe -Fbin -o game_raccoon_boot_image.bin entry_point.68k.asm')
      exit(0)
         
   # Create a faux full 4Mbyte boot partition binary by combining the bare Game Raccoon
   # boot ROM image with an imitation directory cache (created by imitate_directory_cache.py).
   # You need this to test the Game Raccoon boot ROM on an emulator since there wouldn't
   # be any directory cache otherwise.
   # $000000: Mega Drive header
   # $000200: Included binaries
   # $xxxxxx: Entry point
   # $yyyyyy: Padding $00 bytes up to $1FFFFF
   # $200000: Directory cache
   # $zzzzzz: Padding $FF bytes up to $3FFFFF
   if stage == "append_imitation_directory_cache_binary":
      b = open("game_raccoon_boot_image.bin", "rb").read()
      b_appended = b + (b'\x00' * (0x200000 - len(b))) + open("directory_cache_emulator.bin", "rb").read()
      b_appended = b_appended + (b'\xFF' * (0x400000 - len(b_appended)))
      open("game_raccoon_boot_image_directory_cache.bin", "wb").write(b_appended)
      exit(0)
   
   # All of the above targets.
   # End result is game_raccoon_boot_image_directory_cache.bin
   if stage == "complete_boot_rom_image":
      stages = [
         "generate_included_binaries_blob",
         "compile_ramresident",
         "combine_entry_point_ramresident",
         "append_imitation_directory_cache_binary"
      ]
      
      for s in stages:
         result = os.system("python fruitsalad.py %s" % s)
         if result != 0:
            raise Exception(result)
            
      exit(0)
   
   print("Unrecognised stage?")
   exit(-1)

   # Now we can compile the bootstrap which includes the .